

////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                        //
//                                         BRANCH-AND-BOUND                                               //
//                                                                                                        //
////////////////////////////////////////////////////////////////////////////////////////////////////////////


function branch_and_bound () {

//////////////////////////////////////////////////////
//                   VARIABLES                      //
//////////////////////////////////////////////////////

// branch-and-bound variables
var UB = 0; // upper bound
var x_opt = new Object(); // optimal solution
var Y = new Array(); // list of boxes

// control variables
var step = 0; // iteration counter
var run = 0; // start/stop switch: 0 = stop branch-and-bound, 1 = run branch-and-bound
var speed = 1; // fast/slow mode switch: 0 = slow mode, 1 = fast mode
var render_boxes = 1; // render boxes switch: 0 = do not render boxes, 1 = render boxes

// timer variables
var time_last_frame = 0; // timestamp last frame
var render_time = 0; // time needed to render all boxes

// canvas variables
var canvas = document.getElementById('canvas_B');
var context = canvas.getContext('2d');

// mouse over button variable
var mouse_over = 0;



//////////////////////////////////////////////////////
//               CALIBRATEABLE DATA                 //
//////////////////////////////////////////////////////

// fast / slow mode time values
var ms_per_frame = 50; // milliseconds per frame in fast mode
var ms_per_step = 500; // milliseconds between to frames in slow mode

// button (x, y, w, h): button center = (canvas.width/2 + x, canvas.heigth + y), w = width, h = height
var button = new Array(
  new Array(-102, -18, 24, 24),
  new Array(-42, -18, 24, 24),
  new Array(30, -18, 48, 24),
  new Array(102, -18, 24, 24)
);



//////////////////////////////////////////////////////
//             CLASSES AND FUNCTIONS                //
//////////////////////////////////////////////////////


function box (I) {
  this.I = I;
  this.LB = LB(I);
}


function r_default (I) {
  if (typeof window.r == 'function') return r(I);

  if (dimension == 1) return 0.5*(I[0]+I[1]);

  var h = new Array(dimension);
  for (var k = 0; k < dimension; k++) h[k] = 0.5*(I[k][0]+I[k][1]);
  return h;
}


//////////////////////////////////////////////////////
//                     EVENTS                       //
//////////////////////////////////////////////////////


// init branch-and-bound
init_branch_and_bound();

// loop action
window.setInterval(loop, ms_per_frame);



//////////////////////////////////////////////////////
//                   MOUSE ACTION                   //
//////////////////////////////////////////////////////


document.onmousemove = function mouse_move (e) {
  var mouse_x = e.pageX - canvas.offsetLeft;
  var mouse_y = e.pageY - canvas.offsetTop;
  mouse_over = 0;

  for (var k = 0; k < button.length; k++)
  if ((Math.abs(canvas.width/2 + button[k][0] - mouse_x) < button[k][2]/2) && (Math.abs(canvas.height + button[k][1] - mouse_y) < button[k][3]/2)) mouse_over = k+1;

  document.body.style.cursor = (mouse_over == 0 ? 'default' : 'pointer');
  plot_control();
}

  
document.onmouseup = function mouse_up () {
  if (mouse_over == 1) {
    run = (run+1) % 2;
  } else if (mouse_over == 2) {
    init_branch_and_bound();
  } else if (mouse_over == 3) {
    speed = (speed+1) % 2;
  } else if (mouse_over == 4) {
    render_boxes = (render_boxes+1) % 2;
    plot();
  };
  plot_control();
}



//////////////////////////////////////////////////////
//           GEOMETRIC BRANCH-AND-BOUND             //
//////////////////////////////////////////////////////


function init_branch_and_bound () {
  time_last_frame = 0;

  canvas.width = 512;
  canvas.height = dimension * 70 + 132;

  if (typeof window.init_problem == 'function') init_problem();
  Y = new Array();

  if (dimension == 1) {
    Y.push(new box(new Array(X[0], X[1])));
  } else {
    var H = new Array(dimension);
    for (var k = 0; k < dimension; k++) H[k] = new Array(X[k][0], X[k][1]);
    Y.push(new box(H));
  }

  UB = f(r_default(Y[0].I));
  x_opt = r_default(Y[0].I);

  step = 0;
  run = 0;

  plot();
}


function update_UB () {
  var objective_value = f(r_default(Y[Y.length-1].I));
  if (objective_value < UB) {
    UB = objective_value;
    x_opt = r_default(Y[Y.length-1].I);
  }
}


// bisect box perpendicular to the direction of the maximum width component in two subboxes
function split_box_max_width () {
  var H = new Array(dimension);

  var max_width_index = 0;
  for (var k = 1; k < dimension; k++) if (Y[0].I[k][1] - Y[0].I[k][0] > Y[0].I[max_width_index][1] - Y[0].I[max_width_index][0]) max_width_index = k;
  var c = 0.5 * (Y[0].I[max_width_index][0] + Y[0].I[max_width_index][1]);

  H = new Array(dimension);
  for (var k = 0; k < dimension; k++) H[k] = new Array(Y[0].I[k][0], Y[0].I[k][1]);
  H[max_width_index][1] = c;
  Y.push(new box(H));
  update_UB();

  H = new Array(dimension);
  for (var k = 0; k < dimension; k++) H[k] = new Array(Y[0].I[k][0], Y[0].I[k][1]);
  H[max_width_index][0] = c;
  Y.push(new box(H));
  update_UB();

  Y.splice(0, 1);
}


// split box into 2^dimension congruent subboxes
function split_box_all () {
  var H = new Array(dimension);
  var C = new Array(dimension);

  for (var k = 0; k < dimension; k++) C[k] = 0.5 * (Y[0].I[k][0] + Y[0].I[k][1]);

  for (var i = 0; i < Math.pow(2, dimension); i++) 
    for (var k = 0; k < dimension; k++) {
      H = new Array(dimension);
      for (var k = 0; k < dimension; k++) H[k] = new Array(Y[0].I[k][0], Y[0].I[k][1]);
      for (var k = 0; k < dimension; k++) H[k][(i >> k) & 1] = C[k];
      Y.push(new box(H));
      update_UB();
    }

  Y.splice(0, 1);
}


// split box in the one-dimensional case
function split_box_one_dimension () {
  var c = 0.5 * (Y[0].I[0] + Y[0].I[1]);
  Y.push(new box(new Array(Y[0].I[0], c)));
  update_UB();
  Y.push(new box(new Array(c, Y[0].I[1])));
  update_UB();
  Y.splice(0, 1);
}


// box with larges diameter is always Y[0], i.e., at the beginning of the list Y
function split_box () {
  if (dimension == 1) split_box_one_dimension(); else
  if (typeof window.split_option == 'number') {
    if (split_option == 1) split_box_all(); else split_box_max_width();
  } else {
    if (dimension <= 3) split_box_all(); else split_box_max_width();
  }
}


function branch_and_bound_step () {
  if (Y.length > 0) {
    split_box();
    step++;
  }

  var k = 0;
  while (k < Y.length) {
    if (Y[k].LB + epsilon > UB) Y.splice(k, 1); else k++;
  }
}



//////////////////////////////////////////////////////
//                      PLOT                        //
//////////////////////////////////////////////////////


// number of decimal places dependent on an interval I
function find_toFixed (I) {
  if (I[1]-I[0] >= 10000) return 0;
  if (I[1]-I[0] >= 1000) return 1;
  if (I[1]-I[0] >= 100) return 2;
  if (I[1]-I[0] >= 10) return 3;
  if (I[1]-I[0] >= 1) return 4;
  if (I[1]-I[0] >= 0.1) return 5;
  if (I[1]-I[0] >= 0.001) return 6;
  return 8;
}


function plot_triangle (a1, a2, b1, b2, c1, c2) {
  context.beginPath();
  context.moveTo(a1, a2);
  context.lineTo(b1, b2);
  context.lineTo(c1, c2);
  context.closePath();
  context.fill();
}


function plot_caption () {
  var y_offset = 40;

  // format font
  context.textBaseline = 'middle';
  context.font = '12px sans-serif';

  for (var i = 0; i < dimension; i++) {
    // local variable
    var X_I = (dimension == 1 ? X : X[i]);
    var X_min = (dimension == 1 ? X[0] : X[i][0]);
    var X_max = (dimension == 1 ? X[1] : X[i][1]);
    var x_i = (dimension == 1 ? x_opt : x_opt[i]);

    // draw text background
    context.fillStyle = '#E0E0E0';
    context.fillRect(0, y_offset + 10 + 70 * i, canvas.width, 20);
    context.fillStyle = '#000000';

    // draw arrow x_i left
    plot_triangle(6, y_offset + 16 + 70 * i, 6, y_offset + 24 + 70 * i, 10, y_offset + 20 + 70 * i);

    // draw arrow x_i right
    plot_triangle(canvas.width-6, y_offset + 16 + 70 * i, canvas.width-6, y_offset + 24 + 70 * i, canvas.width-10, y_offset + 20 + 70 * i);

    // print limit x_i left
    context.textAlign = 'left';
    context.fillText(X_min.toFixed(find_toFixed(X_I)), 18, y_offset + 20 + 70 * i);

    // print limit x_i right
    context.textAlign = 'right';
    context.fillText(X_max.toFixed(find_toFixed(X_I)), canvas.width-18, y_offset + 20 + 70 * i);

    // print problem solution
    context.textAlign = 'left';
    context.fillText("x", canvas.width/2-45, y_offset + 20 + 70 * i);
    context.fillText("= " + (x_i).toFixed(8), canvas.width/2-30, y_offset + 20 + 70 * i);
    context.font = '8px sans-serif';
    context.fillText((i+1).toFixed(0), canvas.width/2-38, y_offset + 22 + 70 * i);
    context.font = '12px sans-serif';
    if (Y.length == 0) context.fillText("*", canvas.width/2-38, y_offset + 18 + 70 * i);

    // draw problem solution
    if (Y.length == 0) {
      context.beginPath();
      context.arc(canvas.width * ((x_i-X_min)/(X_max-X_min)), y_offset + 30 + 70 * i + 18, 2.5, 0, 2*Math.PI, false);
      context.closePath();
      context.fill();
    }
  }
}


function plot_boxes () {
  var X_width = 0;
  var y_offset = 70;

  for (var i = 0; i < dimension; i++) {
    // local variable
    var X_min = (dimension == 1 ? X[0] : X[i][0]);
    var X_width =  (dimension == 1 ? (X[1]-X[0]) : (X[i][1]-X[i][0]));

    // draw boxes background
    context.fillStyle = '#EAEAEA';
    context.fillRect(0, y_offset + 70 * i, canvas.width, 36);

    // draw boxes
    if (render_boxes == 1) {
      if (i % 4 == 0) context.fillStyle = '#DE102D'; else
      if (i % 4 == 1) context.fillStyle = '#0059C5'; else
      if (i % 4 == 2) context.fillStyle = '#359F2F'; else
      if (i % 4 == 3) context.fillStyle = '#FEC806';
      for (var k = 0; k < Y.length; k++) {
        var Y_min = (dimension == 1 ? Y[k].I[0] : Y[k].I[i][0]);
        var Y_max = (dimension == 1 ? Y[k].I[1] : Y[k].I[i][1]);
        context.fillRect(canvas.width * ((Y_min-X_min)/X_width) - 0.5, y_offset + 70 * i, canvas.width * ((Y_max-Y_min)/X_width) + 1, 36);
      }
    }
  }
}


function plot_result () {
  var y_offset = canvas.height - 75;
  context.fillStyle = '#000000';
  context.font = '12px sans-serif';
  context.textBaseline = 'middle';

  // print upper bound
  context.textAlign = 'right';
  context.fillText("upper bound = ", 140, y_offset);
  context.textAlign = 'left';
  context.fillText(UB.toFixed(10), 140, y_offset);

  // print epsilon
  context.textAlign = 'left';
  context.fillText(epsilon, 140, y_offset + 20);
  context.textAlign = 'right';
  context.fillText("epsilon = ", 140, y_offset + 20);

  // print iteration counter
  context.textAlign = 'left';
  context.fillText(step, canvas.width-110, y_offset);
  context.textAlign = 'right';
  context.fillText("iteration counter = ", canvas.width-110, y_offset);

  // print number of boxes
  context.textAlign = 'left';
  context.fillText(Y.length, canvas.width-110, y_offset + 20);
  context.textAlign = 'right';
  context.fillText("current number of boxes = ", canvas.width-110, y_offset + 20);

  // print render time
  context.fillStyle = '#C0C0C0';
  context.textAlign = 'right';
  context.font = '9px sans-serif';
  context.save();
  context.translate(canvas.width-18, y_offset-14);
  context.rotate(-Math.PI/2);
  context.fillText("render time", 0, 0);
  context.fillText(render_time + " ms", -10, 10);
  context.restore();
}


function plot_head () {
  context.fillStyle = '#222222';
  context.fillRect(0, 0, canvas.width, 40);
  context.textBaseline = 'middle';
  context.textAlign = 'center';
  context.font = 'bold 18px sans-serif';
  context.fillStyle = '#808080';
  context.fillText("Geometric Branch-and-Bound", canvas.width/2, 20);
}


function plot_start_stop () {
  var x_offset = canvas.width / 2 + button[0][0] - button[0][2]/2;
  var y_offset = canvas.height + button[0][1] - button[0][3]/2;
  context.fillStyle = (mouse_over == 1 ? '#F6F6F6' : '#808080');
  if (run == 1) {
    context.fillRect(x_offset + 4, y_offset + 2, 6, 20);
    context.fillRect(x_offset + 14, y_offset + 2, 6, 20);
  } else {
    context.beginPath();
    context.moveTo(x_offset + 3, y_offset + 2);
    context.lineTo(x_offset + 3, y_offset + 22);
    context.lineTo(x_offset + 21, y_offset + 12);
    context.closePath();
    context.fill();
  }
}


function plot_reset () {
  var x_offset = canvas.width / 2 + button[1][0] - button[1][2]/2;
  var y_offset = canvas.height + button[1][1] - button[1][3]/2;
  context.fillStyle = (mouse_over == 2 ? '#F6F6F6' : '#808080');
  context.fillRect(x_offset + 4, y_offset + 4, 16, 16);
}


function plot_speed () {
  var x_offset = canvas.width / 2 + button[2][0] - button[2][2]/2;
  var y_offset = canvas.height + button[2][1] - button[2][3]/2;
  context.strokeStyle = '#808080';
  context.lineWidth = 16;
  context.lineCap = 'round';
  context.beginPath();
  context.moveTo(x_offset + 8, y_offset + 12);
  context.lineTo(x_offset + 40, y_offset + 12);
  context.stroke();
  context.strokeStyle = '#F6F6F6';
  context.beginPath();
  context.moveTo(x_offset + (speed != 1 ? 20 : 28), y_offset + 12);
  context.lineTo(x_offset + (speed != 1 ? 8 : 40), y_offset + 12);
  context.stroke();
}


function plot_render () {
  var x_offset = canvas.width / 2 + button[3][0] - button[3][2]/2;
  var y_offset = canvas.height + button[3][1] - button[3][3]/2;
  context.fillStyle = (mouse_over == 4 ? '#F6F6F6' : '#808080');
  context.strokeStyle = (mouse_over == 4 ? '#F6F6F6' : '#808080');
  if (render_boxes == 1) {
    context.fillRect(x_offset + 1, y_offset + 2, 22, 4);
    context.fillRect(x_offset + 1, y_offset + 10, 22, 4);
    context.fillRect(x_offset + 1, y_offset + 18, 22, 4);
  } else {
    context.lineWidth = 3.5;
    context.lineCap = 'round';
    context.beginPath();
    context.moveTo(x_offset + 6, y_offset + 18);
    context.lineTo(x_offset + 18, y_offset + 6);
    context.stroke();
    context.beginPath();
    context.arc(x_offset + 12, y_offset + 12, 10, 0, 2 * Math.PI, false);
    context.stroke();
  }
}


function plot_control () {
  context.fillStyle = '#222222';
  context.fillRect(0, canvas.height, canvas.width, -36);

  plot_start_stop();
  plot_reset();
  plot_speed();
  plot_render();
}


function plot () {
  context.fillStyle = '#F6F6F6';
  context.fillRect(0, 0, canvas.width, canvas.height);

  render_time = (new Date()).getTime(); // start stopwatch
  plot_boxes();
  plot_caption();
  render_time = (new Date()).getTime() - render_time; // end stopwatch

  plot_result();
  plot_head();
  plot_control();
}



//////////////////////////////////////////////////////
//                    CONTROL                       //
//////////////////////////////////////////////////////


function loop () {
  if (Y.length == 0) run = 0;

  if (run == 1) {
    if (speed == 1) {
      var time_start_frame = (new Date()).getTime();
      do { branch_and_bound_step(); } while ((new Date()).getTime() - time_start_frame < ms_per_frame + render_time);
      plot();
    } else {
      if ((new Date()).getTime() - time_last_frame >= ms_per_step) {
        branch_and_bound_step();
        time_last_frame = (new Date()).getTime();
        plot();
      }
    }
  }

  plot_control();
  if ((typeof window.plot_problem == 'function') && (document.getElementById('canvas_P'))) plot_problem(x_opt);
}



//////////////////////////////////////////////////////
//              END BRANCH-AND-BOUND                //
//////////////////////////////////////////////////////

};





////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                        //
//                                        INTERVAL ARITMETIC                                              //
//                                                                                                        //
////////////////////////////////////////////////////////////////////////////////////////////////////////////



var IA = new function () {

  function max (a, b) {
    return (a >= b ? a : b);
  }

  function min (a, b) {
    return (a <= b ? a : b);
  }

  function x_in (x, a, b) {
    return ((a <= x) && (x <= b));
  }

  function list_max (L) {
    var h = L[0];
    for (var i = 1; i < L.length; i++) if (L[i] > h) h = L[i];
    return h;
  }

  function list_min (L) {
    var h = L[0];
    for (var i = 1; i < L.length; i++) if (L[i] < h) h = L[i];
    return h;
  }

  this.add = function(I, J) {
    return new Array(I[0]+J[0], I[1]+J[1]);
  }

  this.add_scalar = function(a, I) {
    return new Array(a+I[0], a+I[1]);
  }

  this.sub = function (I, J) {
    return new Array(I[0]-J[1], I[1]-J[0]);
  }

  this.mult = function (I, J) {
    var h = new Array(I[0]*J[0], I[0]*J[1], I[1]*J[0], I[1]*J[1]);
    return new Array(list_min(h), list_max(h));
  }

  this.mult_scalar = function (a, I) {
    var h = new Array(a*I[0], a*I[1]);
    return new Array(list_min(h), list_max(h));
  }

  this.div = function (I, J) {
    if ((J[0] <= 0) && (J[1] >= 0)) return new Array(0,0);
    var h = new Array(I[0]/J[0], I[0]/J[1], I[1]/J[0], I[1]/J[1]);
    return new Array(list_min(h), list_max(h));
  }

  this.min = function (I, J) {
    return new Array(min(I[0],J[0]), min(I[1],J[1]));
  }

  this.max = function (I, J) {
    return new Array(max(I[0],J[0]), max(I[1],J[1]));
  }

  this.sq = function (I) {
    if (I[0] >= 0) return new Array(I[0]*I[0], I[1]*I[1]);
    if (I[1] <= 0) return new Array(I[1]*I[1], I[0]*I[0]);
    return new Array(0, max(I[0]*I[0], I[1]*I[1]));
  }

  this.pow = function (I, a) {
    var p = Math.floor(a);
    if (p <= 0) return new Array(1, 1);
    if (p == 1) return new Array(I[0], I[1]);
    if (p == 2) return this.sq(I);
    if (p % 2 == 1) return new Array(Math.pow(I[0],p), Math.pow(I[1],p));
    if (I[0] >= 0) return new Array(Math.pow(I[0],p), Math.pow(I[1],p));
    if (I[1] <= 0) return new Array(Math.pow(I[1],p), Math.pow(I[0],p));
    return new Array(0, max(Math.pow(I[0],p), Math.pow(I[1],p)));
  }

  this.exp = function (I) {
    return new Array(Math.exp(I[0]), Math.exp(I[1]));
  }

  this.atan = function (I) {
    return new Array(Math.atan(I[0]), Math.atan(I[1]));
  }

  this.log = function (I) {
    if (I[0] <= 0) return new Array(0, 0);
    return new Array(Math.log(I[0]), Math.log(I[1]));
  }

  this.abs = function (I) {
    if (I[0] >= 0) return new Array(I[0], I[1]);
    if (I[1] <= 0) return new Array(-I[1], -I[0]);
    return new Array(0, max(-I[0], I[1]));
  }

  this.sqrt = function (I) {
    if (I[1] < 0) return new Array(0, 0);
    if (I[0] < 0) return new Array(0, Math.sqrt(I[1]));
    return new Array(Math.sqrt(I[0]), Math.sqrt(I[1]));
  }

  this.cos = function (I) {
    if (I[1] - I[0] >= 2*Math.PI) return new Array(-1, 1);
    var a = (I[0] >= 0 ? I[0] : I[0] + 2*Math.PI*Math.floor(-I[0]+2) ) % (2*Math.PI); // a is element of [0, 2*PI]
    var b = a + (I[1] - I[0]); // b is element of [0, 4*PI]
    var h = new Array(Math.cos(I[0]), Math.cos(I[1]));
    if ((x_in(  Math.PI, a, b)) || (x_in(3*Math.PI, a, b))) h.push(-1);
    if  (x_in(2*Math.PI, a, b)) h.push(1);
    return new Array(list_min(h), list_max(h));
  }

  this.sin = function (I) {
    if (I[1] - I[0] >= 2*Math.PI) return new Array(-1, 1);
    var a = (I[0] >= 0 ? I[0] : I[0] + 2*Math.PI*Math.floor(-I[0]+2) ) % (2*Math.PI); // a is element of [0, 2*PI]
    var b = a + (I[1] - I[0]); // b is element of [0, 4*PI]
    var h = new Array(Math.sin(I[0]), Math.sin(I[1]));
    if ((x_in(  Math.PI/2, a, b)) || (x_in(5*Math.PI/2, a, b))) h.push(1);
    if ((x_in(3*Math.PI/2, a, b)) || (x_in(7*Math.PI/2, a, b))) h.push(-1);
    return new Array(list_min(h), list_max(h));
  }

};



////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                        //
//                                       PROBLEM VISUALIZATION                                            //
//                                                                                                        //
////////////////////////////////////////////////////////////////////////////////////////////////////////////



var PV = new function () {

  // init variable
  var initialized = false;

  // canvas variables
  var canvas = new Object();
  var context = new Object();

  // relative canvas size
  var x_min = -256;
  var x_max = 256;
  var y_min = -256;
  var y_max = 256;
  var x_ppu = 1;
  var y_ppu = 1;


  function init () {
    canvas = document.getElementById('canvas_P');
    context = canvas.getContext('2d');
    initialized = true;
  }

  this.set_size = function (xmin, xmax, ymin, ymax, ratio) {
    if (initialized == false) init();
    canvas.width = 512;
    canvas.height = Math.floor(ratio*canvas.width) + 40;
    x_min = xmin;
    x_max = xmax;
    y_min = ymin;
    y_max = ymax;
    x_ppu = canvas.width / (x_max - x_min);
    y_ppu = (canvas.height-40) / (y_max - y_min);
    context.fillStyle = '#F6F6F6';
    context.fillRect(0, 0, canvas.width, canvas.height);
  }

  this.head = function (caption) {
    if (initialized == false) init();
    context.fillStyle = '#222222';
    context.fillRect(0, 0, canvas.width, 40);
    context.textBaseline = 'middle';
    context.textAlign = 'center';
    context.fillStyle = '#808080';
    context.font = 'bold 18px sans-serif';
    context.fillText(caption, canvas.width/2, 20);
  }

  this.disc = function (x, y, r, c) {
    if (initialized == false) init();
    context.fillStyle = c;
    context.beginPath();
    context.arc(x_ppu*(x-x_min), canvas.height - y_ppu*(y-y_min), r*x_ppu, 0, 2*Math.PI, false);
    context.closePath();
    context.fill();
  }

  this.dot = function (x, y) {
    if (initialized == false) init();
    context.fillStyle = '#000000';
    context.beginPath();
    context.arc(x_ppu*(x-x_min), canvas.height - y_ppu*(y-y_min), 2.5, 0, 2*Math.PI, false);
    context.closePath();
    context.fill();
  }

  this.circle = function (x, y, r, c) {
    if (initialized == false) init();
    context.strokeStyle = c;
    context.beginPath();
    context.arc(x_ppu*(x-x_min), canvas.height - y_ppu*(y-y_min), r*x_ppu, 0, 2*Math.PI, false);
    context.closePath();
    context.stroke();
  }

  this.line = function (x1, y1, x2, y2, c) {
    if (initialized == false) init();
    context.strokeStyle = c;
    context.beginPath();
    context.moveTo(x_ppu*(x1-x_min), canvas.height - y_ppu*(y1-y_min));
    context.lineTo(x_ppu*(x2-x_min), canvas.height - y_ppu*(y2-y_min));
    context.closePath();
    context.stroke();
  }

  this.text = function (x, y, caption) {
    if (initialized == false) init();
    context.font = '12px sans-serif';
    context.textBaseline = 'middle';
    context.textAlign = 'center';
    context.globalAlpha = 0.75;
    context.fillStyle = '#FFFFFF';
    var metrics_width = context.measureText(caption).width;
    context.fillRect(x_ppu*(x-x_min) - metrics_width/2 - 6, canvas.height - y_ppu*(y-y_min) - 10, metrics_width + 12, 20);
    context.globalAlpha = 1.0;
    context.fillStyle = '#000000';
    context.fillText(caption, x_ppu*(x-x_min), canvas.height - y_ppu*(y-y_min));
  }

  this.rect = function (x, y, w, h, c) {
    if (initialized == false) init();
    context.fillStyle = c;
    context.fillRect(x_ppu*(x-x_min), canvas.height - y_ppu*(y-y_min), w*x_ppu, -h*y_ppu);
  }


};

